#!/usr/bin/env python

from umnet_scripts.utils import *
from umnet_scripts import UMnetequip
import argparse, os
import logging
from pprint import pprint
import re

logger = logging.getLogger(os.path.basename(__file__))

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Audit equipdb against rancid files on wallace-new.")
    parser.add_argument('-l', '--log-level', dest='log_level', default="error",
            help='Set logging level (default=error)')

    args = parser.parse_args()

    logging.basicConfig(
        format='%(asctime)s %(levelname)s %(name)s %(message)s',
        datefmt='%Y-%m-%dT%H:%M:%S%z',
        level=args.log_level.upper(),
    )

    # Get all active switches and routers
    eqdb = UMnetequip()
    devices = eqdb.get_devices_by_category('SWITCH OR ROUTER', active_only=True)

    # first need to merge stack entries in the DB. Sometimes 'monitored device'
    # is checked on one stack member, but 'rancid' is checked on another (ugh)
    merged_devices = {}
    for d in devices:
        if d['dns_name'] in merged_devices.keys():
            curr_d = merged_devices[d['dns_name']]
            curr_d['rancid'] = d['rancid'] or curr_d['rancid']
            curr_d['monitored_device'] = d['monitored_device'] or curr_d['monitored_device']
        else:
            merged_devices[d['dns_name']] = dict(d)

    # we only care about devices that are 'monitored' and 'online'
    monitored_devices = [d for d in merged_devices.values() if d['monitored_device'] and not d['off_line']]

    # Get a list of all the files on this machine's /home/rancid/*configs* directories
    # Stolen from Ben Meekhof in the umnet_netbox_etl code
    rancid_cmd = '''cd /home/rancid && find . -path ./.ssh -prune -false -o -path '*/configs/*' \
            -not -path '*CVS*'  -not -path './RouteTables/*' \
            -not -path './ARP/*' -not -path './CiscoWL/*' \
            -not -path './TEST/*' -mtime -7 -type f \
    '''
    with os.popen(rancid_cmd) as stream:
        lines = stream.read().splitlines()

    # peel off filename which maps to dns_name or IP address
    # of the device
    rancid_devices = [l.split('/')[-1].replace('.umnet.umich.edu','') for l in lines]

    # First let's print out all the 'on-line' devices in equipdb that don't have the rancid box checked
    no_rancid = [d['dns_name'] for d in monitored_devices if not d['rancid']] 
    if(no_rancid):
        print("The following devices are not marked as 'offline' but don't have their 'rancid' checkbox checked:\n")
        [print(d) for d in no_rancid]
    else:
        print("All online 'SWITCH_OR_ROUTER' devices are in rancid.")

    print("\n")
    # Now we'll print all the devices marked for rancid that don't have a recent backup
    failed_rancid = [d['dns_name'] for d in monitored_devices
        if d['dns_name'] not in rancid_devices and
           d['ip_address'] not in rancid_devices and
           d['rancid']
           ]

    if(failed_rancid):
        print("The following devices are checked for rancid backup but haven't been backed up recenty.\n")
        [print(d) for d in failed_rancid]
    else:
        print("All devices passed rancid backup check")



from sqlalchemy import create_engine
from .utils import get_env_vars
from .umnetdb import UMnetdb
import logging

logger = logging.getLogger(__name__)

class UMnetequip(UMnetdb):
    '''
    This class wraps helpful equip db queries in python.
    The API is lame.
    '''
    def __init__(self, host='mallard.dsc.umich.edu', port=1521):

        eqdb_creds = get_env_vars(['EQUIP_DB_USERNAME','EQUIP_DB_PASSWORD'])
        self._url = f"oracle://{eqdb_creds['EQUIP_DB_USERNAME']}:{eqdb_creds['EQUIP_DB_PASSWORD']}@(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))(CONNECT_DATA=(SID=PC13)))"
        self._e = create_engine(self._url)
        
        logger.debug(f"Created DB engine {self._url}")


    def get_devices_by_category(self, category, active_only=False):
        '''
        Queries equip db for devices by category. You can also
        specify if you only want active devices.
        '''

        select = [ 'eq.monitored_device', 
                   'eq.rancid',
                   'eq.off_line',
                   'eq.dns_name',
                   'eq.ip_address',
                 ]
        table = 'ARADMIN1.MAC_SLA_EQUIPMENT eq'

        where = [f"eq.category = '{category}'"]

        # Equipshim status numbers (see ARADMIN1.UMNET_EQUIPMENT_STATUS)
        # 1: RESERVE, 2:ACTIVE, 3:RETIRED
        if active_only:
            where.append("eq.status = 2")
        
        sql = self._build_select(select, table, where=where, distinct=True)
        return self._execute(sql)



from enum import Enum
from ncclient import manager
import logging
import json
from lxml import etree
from .constants import NETBOX_NSO_NEDS
from datetime import datetime
from pprint import pprint

NCS_NAMESPACE = 'http://tail-f.com/ns/ncs'
NAMESPACES = {'ncs':NCS_NAMESPACE}

logger = logging.getLogger()


class UMnetnso:

    def __init__(self, host='localhost', port=2022, 
            user='admin', password='admin'):

        self.m = manager.connect_ssh(
                host,
                username=user,
                password=password,
                port=port,
                look_for_keys=False,
                hostkey_verify=False
                )

        self._device_groups = None

    def _get_xpath(self, get_type, xpath, filter_on=[], source='running', with_defaults=None):
        '''
        Builds an xpath get or get_config query based on a provided absolute path
        and an optional list of things to filter on at that path. Then calls the ncclient get or get_config
        method and returns a result.

        The 'filtered on' input is for filtering out the child nodes retrieved at that path and
        will get 'ored' together.

        Eg: filter_on = ['address','port'] will get translated to '*[self::address|self::port]'

        xpath :str: a string representing an absolute path
        get_type :str: either 'get' or 'get_config'
        filter_on :list: a list of nodes to further filter for
            at the provided path - they will get ORed together.
        source :string: The source datastore (eg "running" or "candidate"), only
            applicable for get_config
        with_defaults :string: see RFC6243, tells NSO whether to return default
            config values or not
        Result is returned as :class: xml.etree.ElementTree.Element
        '''

        if filter_on:
            filter_str = "|".join([f"self::{f}" for f in filter_on])
            filtered_xpath = f'{xpath}*[{filter_str}]'
        else:
            filtered_xpath = xpath

        logger.debug(f"_get_xpath: generated xpath {filtered_xpath}")

        if get_type == 'get':
            result = self.m.get(filter=('xpath', filtered_xpath), with_defaults=with_defaults)

        elif get_type == 'get_config':
            result = self.m.get_config(source=source, filter=('xpath', filtered_xpath), with_defaults=with_defaults)

        else:
            raise ValueError(f"Invalid netconf get type {get_type}")

        return result.data_ele

    def get_device_groups(self, device_name=None, force_query=False):
        '''
        Returns a list of nso group names. If you provide a device name
        it will return the groups tied only to that device.

        The function sets an internal object called _device_group when
        first queried. If you don't set force_query to "true", it will
        use that cached object for subsequent calls.

        device_name :str: name of device
        force_query :bool: Force a new call to nso 
        '''

        if self._device_groups == None or force_query:
            self._device_groups = self._get_xpath('get_config', '/devices/device-group')
        
        if device_name:
            query = f"//ncs:device-group[ncs:device-name='{device_name}']/ncs:name/text()"
        else:
            query = "//ncs:device-group/ncs:name/text()"
       
        return self._device_groups.xpath(query, namespaces=NAMESPACES)



    def get_nso_devices(self):
        '''
        Gets a subset of basic config info for all devices in NSO. Returns results
        as a list of dicts with the following format:
            [{'name':name, 'address':address, 'ned-id', ned-id, 'admin-state':admin-state,
               'groups':[list of groups] } ]
        '''

        # Get device addresses, type (NED), and admin state. We'll also get the device
        # name since it's the key for the devices/device leaflist.
        xpath='/devices/device/'
        filter_on = ['device-type','address','state']
        result = self._get_xpath('get_config', xpath, filter_on=filter_on, with_defaults='report-all')

        # extract values from leafs
        leafs = ['name','address','admin-state','ned-id']
        output_lists = [result.xpath(f'//ncs:{l}/text()',namespaces=NAMESPACES) for l in leafs]

        
        # this transforms a list of lists [ [name1, name2,...], [ip1, ip2, ...] ...]
        # into a list of dicts
        #  [{'name':name1,'address':ip1,...}, {'name':name2,'address':ip2,...}... ]
        output = [ dict(zip(leafs,device)) for device in list(zip(*output_lists)) ]
       
        # Add group membership to device entries
        for d in output:
            d['groups'] = set(self.get_device_groups(device_name=d['name']))

        return output

    def update_nb_devices(self, nb_devices, operation='merge'):
        '''
        This takes a list of pynetbox device objects and updates the NSO device
        and device group leaflists
        :devices: list of pynetbox device objects
        :operation: the netconf operation for the devices (default: merge)
        '''
        # First we need a current list of the device group assignments from NSO.
        xpath='/devices/device-group'
        nso_groups = self._get_xpath('get_config', xpath)

        # Now let's start building our tree
        config = etree.Element("{urn:ietf:params:xml:ns:netconf:base:1.0}config")
        devices = etree.SubElement(config, "devices", nsmap={None:NCS_NAMESPACE})

        # add elements
        for d in nb_devices:

            e_device = etree.SubElement(devices, "device",operation=operation)
            etree.SubElement(e_device, "name").text = d.name
            etree.SubElement(e_device, "address").text = d.primary_ip4.address.split("/")[0]
            etree.SubElement(e_device, "port").text = "22"
            etree.SubElement(e_device, "authgroup").text = "local"
            
            try:
                ned = NETBOX_NSO_NEDS[d.platform.name]
            except KeyError:
                logger.error("ERROR: no NED for {d.platform.name}, skipping")
                continue

            d_type = etree.SubElement(e_device, "device-type")
            ned_type = etree.SubElement(d_type, ned["device-type"])
            etree.SubElement(ned_type, "ned-id").text = ned["ned-id"]

            source = etree.SubElement(e_device, "source")
            etree.SubElement(source, "when").text = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            etree.SubElement(source, "source").text = "Netbox audit script"

            # current device groups are based on site and device role
            nb_groups = [ d.device_role.slug, d.site.slug ]

            # pull existing groups for this device
            curr_groups = self.get_device_groups(d.name)

            # if a device is in a group it shouldn't be, remove that.
            for g in [g for g in curr_groups if g not in nb_groups]:
                e_group = etree.SubElement(devices, "device-group")
                etree.SubElement(e_group, "name").text = g
                etree.SubElement(e_group, "device-name",operation='delete').text = d.name

            # if a device needs to be added to a group, add it
            for g in [g for g in nb_groups if g not in curr_groups]:
                e_group = etree.SubElement(devices, "device-group")
                etree.SubElement(e_group, "name").text = g
                etree.SubElement(e_group, "device-name").text = d.name

        result = self.m.edit_config(config, target="running")



        return result

    def remove_devices(self, device_names):
        '''
        This takes a list of device names and removes them from NSO.
        :device_names: list of device names
        '''

        config = etree.Element("{urn:ietf:params:xml:ns:netconf:base:1.0}config")
        devices = etree.SubElement(config, "devices", nsmap={None:'http://tail-f.com/ns/ncs'})

        # reset device group internal element
        self._device_groups = None

        for d in device_names:
            e_device = etree.SubElement(devices, "device", operation="delete")
            etree.SubElement(e_device, "name").text = d

            groups = self.get_device_groups(d)
            for g in groups:
                e_group = etree.SubElement(devices,"device-group")
                etree.SubElement(e_group,"name").text = g
                etree.SubElement(e_group,"device-name",operation='delete').text = d

        result = self.m.edit_config(config, target="running")

        return result



from sqlalchemy import create_engine
from .utils import get_env_vars
import logging

logger = logging.getLogger(__name__)

class UMnetdb:
    '''
    This class wraps helpful netinfo db queries in python.
    '''
    def __init__(self):
        '''
        DB initialization is unique enough across our various dbs that this is implemented
        in the child classes (at least for now).
        '''
        raise NotImplementedError


    def _build_select(self, select, table, joins=[], where=[], order_by=None, limit=None, group_by=None, distinct=False):
        '''
        Generic 'select' query string builder built from standard query components as input.
        The user is required to generate substrings for the more complex inputs
        (eg joins, where statements), this function just puts all the components
        together in the right order with appropriately-added newlines (for debugging purposes)
        and returns the result.

        :select: a list of columns to select
          ex: ["nip.mac", "nip.ip", "n.switch", "n.port"]
        :table: a string representing a table (or comma-separated tables, with our without aliases)
          ex: "node_ip nip"
        :joins: a list of strings representing join statements. Include the actual 'join' part!
          ex: ["join node n on nip.mac = n.mac", "join device d on d.ip = n.switch"]
        :where: A list of 'where' statements without the 'where'. The list of statements are
          "anded". If you need "or", embed it in one of your list items
          ex: ["node_ip.ip = '1.2.3.4'", "node.switch = '10.233.0.5'"]
        :order_by: A string representing a column name (or names) to order by
        :group_by: A string representing a column name (or names) to group by
        :limit: An integer

        '''

        # First part of the sql statement is the 'select'
        distinct = 'distinct ' if distinct else ''
        sql = f"select {distinct}" + ", ".join(select) + "\n"

        # Next is the table
        sql += f"from {table}\n"

        # Now are the joins. The 'join' keyword(s) need to be provided
        # as part of the input, allowing for different join types
        for j in joins:
            sql += f"{j}\n"

        # Next are the filters. They are 'anded'
        sql += "where\n"
        sql += " and\n".join(where) + "\n"

        # Finally the other options
        if order_by:
            sql += f"order by {order_by}\n"

        if group_by:
            sql += f"group by {group_by}\n"

        if limit:
            sql += f"limit {limit}\n"
        
        logger.debug(f"Generated SQL command:\n****\n{sql}\n****\n")

        return sql

    def _execute(self, sql):
        '''
        Generic sqlalchemy "execute this sql command and give me all the results"
        '''
        with self._e.begin() as c:
            r = c.execute(sql)
        return r.fetchall()




import os
import logging
from dns import resolver
from .utils import resolve_name_or_ip
from .constants import CORE_IPS, VALID_PORT, RANCID_CONFIG_DIRS

logger = logging.getLogger(__name__)

class Device:

    valid_attrs = {
        'hostname':'',
        'ip':'',
        'rancid_type':'',
        'model':'',
        'cfg_file':'',
        'status':'',
        'neighbors':{},
        }

    def __init__(self, **kwargs):

        for attr in self.valid_attrs:
            if attr in kwargs:
                setattr(self, k, kwargs[k])
            else:
                setattr(self, k, self.valid_attrs[k])

    def __setattr__(self, attr, value):
        if attr not in valid_attrs:
            raise ValueError(f'Invalid attr {attr}: valid attrs are {valid_attrs}')
        self.__dict__[attr] = value

    @property
    def rancid_name(self):
        '''
        The 'rancid name' of the device is the name
        the device is known as - either dns name (core/dl/dc/etc)
        or ip address (al). We're deriving it from the config
        file
        '''
        return os.path.basename(self.cfg_file)

class Rancid:
    '''
    A class that provides easy access to rancid data
    stored on the local filesystem. This class is meant to
    be used on wallace/gromit. To use this succesfully
    on a different system you'll have to manually
    re-create '/home/rancid/' on your local system.
    '''

    def __init__(self, 
                rancid_dir:str="/home/rancid", 
                cfg_dirs:list=RANCID_CONFIG_DIRS,
                active_only:bool=False):
        
        # first we're going to populate an internal device
        # list with Device objects from every router.db
        # in every rancid directory
        self.rancid_dir = rancid_dir
        self.cfg_dirs = cfg_dirs

        self._devices = []
        for d in cfg_dirs:
            routerdb = f'{rancid_dir}/{d}/router.db'
            self._devices.extend(self._parse_rdb(d))

        # Now we're going to parse topologywalker and build connections
        # between the devices
        for d in self._devices:
            iftable_file = f'{self.rancid_dir}/Topology/{d.rancid_name}.ifTable'
            neighs = self._parse_iftable(iftable_file)

            # parse_iftable returns a dict of 
            # tuples neighs[port] = (neigh_ip, neigh_port)
            for port, neigh in neighs.items():
                n_device = [d for d in self._devices if d.ip == neigh[0]]
                if n_device:
                    d.neighbors[port] = n_device[0]


    def _munge_port(self, port):
        '''
        Converts port reported by ifTable to the 'short'
        names we use in NSO
        '''

        # I_ABBR maps cisco short names to long ones
        for short, long in I_ABBR.items():
            if port.startswith(long):
                port.replace(long, short)
        
        # also need to remove ".0" from junos ports
        port = re.sub(r'\.0$','', port)

        return port

    def _parse_rdb(self, db_file:str, active_only:bool=False) -> list:
        '''
        Parses a router db file
        '''

        with open(db_file) as fh:
            rdb = fh.read.splitlines()

        devices = []

        for l in rdb:

            #rdb fields are name:rancid_type:status:platform
            fields = l.split(":")
            if fields[2] == 'up' or not(active_only):
                
                device = Device()
                name_and_ip = resolve_name_or_ip(fields[1])
                device.hostname = name_and_ip['dns']
                device.ip = name_and_ip['ip']

                device.rancid_type = fields[1]
                device.status = fields[2]
                device.model = fields[3]
                device.cfg_file = db_file.replace('router.db',f'configs/{fields[0]}')
            
                devices.append(device)

    def _parse_iftable(self, file):
        '''
        Given a path to a ".ifTable" file on the local
        filesystem, parses the file into a dict of
        on interface name.

        Also shortens/cleans up port names to match their
        well-known abbreviations
        '''

        if not(os.path.exists(file)):
            self.log.error(f'{file} not found!')
            return {}

        with open(file) as fh:
            lines = fh.read().splitlines()

        ports = {}
        for l in lines:
            l = l.strip()
            m = re.match(r'^\d+\s+(\S+)\s+(\d+\.\d+\.\d+\.\d+)\s+(\S+)$',l)
            if m:
                port = self._munge_port(m.group(1))
                neigh_ip = m.group(2)
                neigh_port = self._munge_port(m.group(3))
                if neigh_ip not in CORE_IPS \
                    and VALID_PORT.match(port) \
                    and VALID_PORT.match(neigh_port):
                    ports[port] = (neigh_ip, neigh_port)

        return ports

    def get_device(self, name_or_ip:str) -> Device:
        '''
        returns a device based on name or ip
        '''
        name_or_ip = resolve_name_or_ip(fields[1])
        device = [d for d in self._devices 
                    if d.hostname == name_or_ip['name']
                    or d.ip == name_or_ip['ip']
                ]
        if len(device) == 0:
            return False
        if len(device) == 1:
            return device[0]
        raise LookupError(f'More than one device found for {name_or_ip}')


    def get_devices(self, **kwargs) -> list:
        '''
        Look up a set of devices based on any set of device attributes.
        If no arguments are provided, the full device list is returned.
        '''
        results = self._devices.copy()

        for k,v in kwargs.items():
            results = [ r for r in results if getattr(r,k) == v ]

        return results
 




from sqlalchemy import create_engine
import re
from .utils import get_env_vars, is_ip_address, is_mac_address, expand_interface
import logging
from .umnetdb import UMnetdb

logger = logging.getLogger(__name__)

class Umnetdisco(UMnetdb):
    '''
    This class wraps helpful netdisco db queries in python.
    The API is lame.
    '''
    def __init__(self, host='172.0.0.1', port=5432):

        creds = get_env_vars(['NETDISCO_DB_USER','NETDISCO_DB_PASSWORD'])

        self._url = f"postgresql://{creds['NETDISCO_DB_USER']}:{creds['NETDISCO_DB_PASSWORD']}@{host}:{port}/netdisco"
        self._e = create_engine(self._url)
        
        logger.debug(f"Created DB engine {self._url}")


    def host_arp(self, host, start_time=None, end_time=None, limit=1, active_only=False):
        '''
        Does an ARP query against the netdisco db for a single host. 
        :host: A string representing a MAC address or an IPv4 address
        '''

        # what table are we querying?
        table = 'node_ip nip'
        joins = ['join node n on nip.mac = n.mac']

        # define select statements
        select = ['nip.mac', 'nip.ip', 'n.switch', 'n.port', 'nip.time_first', 'nip.time_last']

        # First determine if this host is a MAC address or an IP
        where = []
        if is_mac_address(host):
            where.append(f"nip.mac ='{host}'")
        elif is_ip_address(host, version=4):
            where.append(f"nip.ip ='{host}'")
        else:
            raise ValueError(f"{host} is not a valid IP or mac address")

        # filter for specific start/end times if specified
        if start_time and end_time:
            where.append(f"n.time_last between timestamp '{start_time}' and timestamp '{end_time}'")
        elif start_time:
            where.append(f"n.time_last > timestamp '{start_time}'")

        # filter for active if defined
        if active_only:
            where.append(f"nip.active = true")

        # order by last seen
        sql = self._build_select(select, table, joins=joins, where=where, order_by="time_last", limit=limit)
        result = self._execute(sql)

        return result

    def arp_count(self, prefix=None, start_time=None, end_time=None, active_only=False):
        '''
        Queries the host data in netdisco based on prefix and start/end time.
        First any prefix greater than or equal to the inputted prefix is searched for
        (in the 'device_ip' table).

        Then the host table is queried.
        '''
        
        # We're counting all the host IPs by subnet
        select = ['count(distinct nip.ip)', 'dip.subnet']
        table = 'node_ip nip'

        # postgres allows us to do a join based on an IPs (type inet) membership
        # of a subnet (type cidr)
        joins = ['join device_ip dip on nip.ip <<= dip.subnet']

        # grouping by subnet is what gives us per-subnet host counts
        group_by = 'dip.subnet'

        # append all cli filtering options
        where = [f"dip.subnet <<= inet '{prefix}'"]
        if start_time and end_time:
            where.append(f"nip.time_last between timestamp '{start_time}' and timestamp '{end_time}'")
        elif start_time:
            where.append(f"nip.time_last > timestamp '{start_time}'")

        if active_only:
            where.append(f"nip.active = true")

        sql = self._build_select(select, table, joins=joins, where=where, group_by=group_by)
        return self._execute(sql)

    def neighbors(self, device=None):
        '''
        Queries the device_ip table in netdisco to get neighbors of a device.
        If no device is specified, all neighbors are pulled.
        The device input can be an IPv4 address or an FQDN.
        '''

        select = [ 'dp.ip as local_ip',
                   'ld.dns as local_dns',
                   'dp.port as local_port',
                   'dp.remote_ip',
                   'rd.dns as remote_dns',
                   'dp.remote_port']
        table = 'device d'
        joins = ['join device_port dp on d.ip = dp.ip',
                 'join device ld on dp.ip = ld.ip',
                 'join device rd on dp.remote_ip = rd.ip']

        where = ['dp.remote_ip is not null']
        if is_ip_address(device):
            where.append(f"d.ip = '{device}'")
        elif device:
            where.append(f"d.dns = '{device}'")

        sql = self._build_select(select, table, joins, where)
        return self._execute(sql)

    def get_devices(self, match_subnets=None, exclude_subnets=None):
        '''
        Queries netdisco for a list of devices.
        Optionally, limit it by a list of prefixes
        '''

        select = [ 'ip', 'dns', 'serial', 'model', 'vendor', 'os' ]
        table = 'device'

        where = []
        if match_subnets:
            where.append(" or\n".join([f"ip << inet '{s}'" for s in match_subnets]))
        
        if exclude_subnets:
            where.append("not\n" + " or\n".join([f"ip << inet '{s}'" for s in exclude_subnets]))

        sql = self._build_select(select, table, where=where)
        return self._execute(sql)



from sqlalchemy import create_engine
from .utils import get_env_vars
from .umnetdb import UMnetdb
import logging
import ipaddress

logger = logging.getLogger(__name__)

class UMnetinfo(UMnetdb):
    '''
    This class wraps helpful netinfo db queries in python.
    '''
    def __init__(self, host='kannada.web.itd.umich.edu', port=1521):

        eqdb_creds = get_env_vars(['NETINFO_USERNAME','NETINFO_PASSWORD'])
        self._url = f"oracle://{eqdb_creds['NETINFO_USERNAME']}:{eqdb_creds['NETINFO_PASSWORD']}@(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))(CONNECT_DATA=(SID=KANNADA)))"
        self._e = create_engine(self._url)
        
        logger.debug(f"Created DB engine {self._url}")


    def get_network_by_ip(self, ip, active_only=False):
        '''
        Looks up a network based on an IPv4 or IPv6 address. Returns the netname,
        vlan id, as well as *all active subnets* (IPv4 and IPv6) tied to the netname.
        '''

        ip = ipaddress.ip_address(ip)

        # to make our lives simpler we're breaking this up into two steps.
        # first let's find the network entry in the 'item' table
        if ip.version == 4:
            select = ["p.itemidnum"]
            table = 'UMNET_ONLINE.IP_SUBNET ip'
            joins = [
                'join UMNET_ONLINE.ITEM i on ip.itemidnum = i.itemidnum',
                'join UMNET_ONLINE.ITEM p on i.parentitemidnum = p.itemidnum',
                ]
            where = [
                    f"{int(ip)} >= ip.ADDRESS32BIT",
                    f"{int(ip)} <= ip.ENDADDRESS32BIT",
                    ]

        # IPv6 table is related to the 'item' table via the RELATED_ITEM table
        elif ip.version == 6:
            select = ["p.itemidnum"]
            table = 'UMNET_ONLINE.IP6NET ip'
            joins = [
                'join UMNET_ONLINE.RELATED_iTEM r on r.itemidnum = ip.itemidnum',
                'join UMNET_ONLINE.ITEM p on p.itemidnum = r.relitemidnum',
                    ]
            # and the start/end addresses are stored as hex strings
            addr_str = ip.exploded.replace(":","")
            where = [
                    f"'{addr_str}' >= ip.ADDRESS128BIT",
                    f"'{addr_str}' <= ip.ENDADDRESS128BIT",
                    ]

        # add status check if user wants it    
        if active_only:
            where.append("p.statuscd='A'")

        sql = self._build_select(select, table, joins=joins, where=where)
        network = self._execute(sql)

        if not(network):
            return False
        net_id = network[0][0]

        # Now let's use the network itemidnum to find all associated subnets
        # with that network, as well as the netname and VLAN ID
        select = [
                "p.itemname as netname",
                "i.itemname as ipv4_subnet",
                "six_i.itemname as ipv6_subnet",
                "NVL(n.vlanid,n.local_vlanid) AS vlan_id",
                ]
        table = "UMNET_ONLINE.ITEM p"
        joins = [
                "join UMNET_ONLINE.NETWORK n on p.itemidnum = n.itemidnum",
                "join UMNET_ONLINE.ITEM i on i.parentitemidnum = p.itemidnum",
                "join UMNET_ONLINE.IP_SUBNET ip on i.itemidnum = ip.itemidnum",
                "join UMNET_ONLINE.RELATED_ITEM r on p.itemidnum = r.relitemidnum",
                "join UMNET_ONLINE.IP6NET six on r.itemidnum = six.itemidnum",
                "join UMNET_ONLINE.ITEM six_i on six.itemidnum = six_i.itemidnum",
                ]
        where = [f"p.itemidnum = {net_id}"]

        sql = self._build_select(select, table, joins=joins, where=where)
        results = self._execute(sql)
        return results



